# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.






# Sampling 

import random
import math
import numpy as np

from .tm_geometry import dirP_to_coord, rotation_matrix, dirC_to_dirP
from scipy.interpolate import interp1d


# Sample a direction from an isotropic distribution 
def sample_Lambertian():
    zenith = math.acos(math.sqrt(random.random())) *180/math.pi
    azimuthal = random.uniform(0,360)
    coord = dirP_to_coord(1,[zenith, azimuthal])
    
    # a list of two lists, coord & direction 
    return [coord, [zenith, azimuthal]] 

    
# Sample a scattering direction based on existing direction 
def sample_scattering(ot_mie,ot_rayleigh,pt_direction,aerosol_SPF, print_on=False): 
    
    # sum of scattering 
    ot_sum = ot_mie + ot_rayleigh

    ### Determine if mie or rayleigh 
    
    ot_random = random.uniform(0,ot_sum)
    
    # Mie 
    if ot_random <= ot_mie:
        if print_on: print ('\nMie scattering')
        type_scat = 'M'
        df_angle = aerosol_SPF.Angle.to_numpy()
        
        # sin correction 
        df_value = aerosol_SPF.Value.to_numpy() * np.sin(df_angle * math.pi/180)
        
        x_min = np.min(df_angle)
        x_max = np.max(df_angle)
        
        if x_min!=0 or x_max!=180:
            print('WARNING: Angle has to be between 0 and 180') # csv problem
 
        f2 = interp1d(df_angle, df_value, kind='cubic') 
        
        '''
        #visualize the interpolation 
        xnew = np.linspace(0, 180, num=800, endpoint=True)
        import matplotlib.pyplot as plt
        plt.plot(df_angle, df_value, 'o', xnew, f(xnew), '-', xnew, f2(xnew), '--')
        plt.legend(['data', 'linear', 'cubic'], loc='best')
        plt.show()
        '''
        
        y_max = np.max(df_value)
        y=y_max
        y_calculated=0
        
        while y>y_calculated:
            y=random.uniform(0,y_max)
            x=random.uniform(0,180)
            y_calculated = f2(x).item() # interpolate 
                
    # Rayleigh   
    else:
        if print_on: print ('\nRayleigh scattering')
        type_scat = 'R'
        
        # # Old method 
        # y_max = 0.82
        # y=y_max
        # y_calculated=0
        
        # while y>y_calculated:
        #     y=random.uniform(0,y_max)
        #     x=random.uniform(0,180)
        #     y_calculated = (3/4)*(1+(math.cos(x/180*math.pi))**2)
        #     y_calculated = y_calculated * math.sin(x/180*math.pi)
        
        # Rayleigh scattering phase function from libRadtran, line 4539 in mystic.c
        P = random.random()
        q = 8.0 * P - 4.0
        u =  (-q / 2.0 + math.sqrt (1.0 + q * q / 4.0))**(1/3)
        v = -1.0 / u
        mu = u + v
        x = math.acos(mu) * 180 / math.pi # sampled scattering angle 
        y_calculated = (3/4)*(1+(math.cos(x/180*math.pi))**2) # scattering intensity 

    # print(x) # azimuthal direction 
    sampled_direction = [x, random.uniform(0,360)]
    
    if print_on: print("Sampled_direction: " + str(sampled_direction))
    
    
    ###### Rest of the function is to rotate the sampled scattering direction towards the existing moving direction 
    
    # find an equator point using the azimuth angle of the sampled direction
    equator_point = [90, sampled_direction[1] + 90] # azimuth plus 90 to find a rotation axis 
    equator_point_C = dirP_to_coord(1,equator_point)
    
    # we rotate this equator point towards pt_direction
    # then it will be perpendicular to pt_direction 

    axis = [math.cos((pt_direction[1]+90)*math.pi/180),
            math.cos(pt_direction[1]*math.pi/180),
            0]     
    
    theta = pt_direction[0]*math.pi/180     
    equator_point_C_rotated = np.dot(rotation_matrix(axis, theta), equator_point_C)
    
    # use the new coordinates as an axis to rotate pt_direction by the scattering angle   
    coord = dirP_to_coord(1,pt_direction)
    axis2 = equator_point_C_rotated
    theta2 = sampled_direction[0]*math.pi/180 #math.pi  # zenith
    rotated = np.dot(rotation_matrix(axis2, theta2), coord)
    new_direction = dirC_to_dirP(rotated)[0:2]
    
    if print_on: print("Rotated_direction: " + str(new_direction)) 
    
    # intensity at new_direction 
    intensity = y_calculated  / math.sin(x/180*math.pi)
    
    if print_on: print ('  intensity: ' + str(intensity))
    
    return new_direction, intensity, type_scat



# Importance sampling 
def weight_impSampling(ot_mie,ot_rayleigh,angle_impSampling,aerosol_SPF, print_on=False): 
    
    # sum of scattering 
    ot_sum = ot_mie + ot_rayleigh
    ot_random = random.uniform(0,ot_sum)
    
    ### Determine if mie or rayleigh 
    
    # Mie 
    if ot_random <= ot_mie:
        if print_on: print ('\nMie scattering importance sampling')
        
        df_angle = aerosol_SPF.Angle.to_numpy()
        df_value = aerosol_SPF.Value.to_numpy() 
        
        x_min = np.min(df_angle)
        x_max = np.max(df_angle)
        
        if x_min!=0 or x_max!=180:
            print('WARNING: Angle has to be between 0 and 180') # csv problem

        # Interpolate 
        f2 = interp1d(df_angle, df_value, kind='cubic') 
        y_calculated = f2(angle_impSampling).item() 
                
    # Rayleigh   
    else:
        if print_on: print ('\nRayleigh scattering importance sampling')
        
        y_calculated = (3/4)*(1+(math.cos(angle_impSampling/180*math.pi))**2)

    # Intensity at new_direction 
    intensity = y_calculated 
    if print_on: print ('  importance sampling intensity: ' + str(intensity))

    return intensity
  
    
# Normalized to a combined SPF model 
def weight_impSampling2(ot_mie,ot_rayleigh,angle_impSampling,aerosol_SPF, print_on=False):  
    
    # sum of scattering 
    ot_sum = ot_mie + ot_rayleigh
    
    # Mie 
    df_angle = aerosol_SPF.Angle.to_numpy()
    df_value = aerosol_SPF.Value.to_numpy() 
    
    x_min = np.min(df_angle)
    x_max = np.max(df_angle)
    
    if x_min!=0 or x_max!=180:
        print('WARNING: Angle has to be between 0 and 180') # csv problem

    f2 = interp1d(df_angle, df_value, kind='cubic') 
    y_calculated_M = f2(angle_impSampling).item() # interpolate 
                
    # Rayleigh   
    y_calculated_R = (3/4)*(1+(math.cos(angle_impSampling/180*math.pi))**2)

    # intensity at new_direction 
    intensity =  y_calculated_M * (ot_mie/ot_sum) + y_calculated_R * (ot_rayleigh/ot_sum)
    
    if print_on: print ('  importance sampling intensity: ' + str(intensity))

    return intensity    
    

# # Not used anymore
# def sample_distance2scatter(ot_aerosol,ot_rayleigh,layer_height, pr):
    
#     ot_sum = ot_aerosol + ot_rayleigh
    

#     if ot_sum > 0: # if there is scattering 
#         if pr: 
#             print ("\nCalculating distance before scattered, layer_height: " + str(layer_height))
        
#         if layer_height is None:
#             print ("WARNING: atmosphere layer-height error ")
        
        
#         # attenuation coefficient due to scattering 
#         c = (ot_aerosol + ot_rayleigh )/layer_height 
        
#         R = random.random()
        
#         pt_linear_distance = - np.log(R) / c # according to Mobley light and water chapter 6
        
#         scatter_at_end = True
        
        
#     else: # if scattering coefficient is 0  
#         if pr:
#             print ('no scattering, pre-set distance')
#         pt_linear_distance = 100_000    
        
#         scatter_at_end = False
        
        
#     if pr:
#         print ('sampled_distance: ' + str(pt_linear_distance))
        
#     return pt_linear_distance, scatter_at_end




# # test sampling_scattering 


# if __name__=='__main__':
#     from Py6S.Params.atmosprofile import AtmosProfile
#     from Atmosphere import Atmosphere 
#     atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) # same as 6S
#     aot550 = 0.1 
#     aerosol_SPF = 'aerosol_maritime_SPF.csv'
#     aerosol_EXT = 'aerosol_maritime_EXT.csv'
#     aerosol_SSA = 'aerosol_maritime_SSA.csv'
    

#     # all wavelengths 
#     my_atm = Atmosphere(atm_profile, aot550, aerosol_SPF, aerosol_EXT, aerosol_SSA)
    
#     # 550
#     my_atm_OT, aerosol_SPF  = my_atm.wavelength(wl=550)
    
#     ot_mie = 0
#     ot_rayleigh = 1
#     pt_direction = [0,0]

#     test = sample_scattering(ot_mie,ot_rayleigh,pt_direction,aerosol_SPF)





















# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.





# Aerosol SPF

import numpy as np
import pandas as pd
import os.path

def find_aerosolSPF(aerosol_type,wl):

    # wl = 3750
    
    # Currently only supporting the default number of angles in 6S
    # For higher angular resolution SPF or mixed aerosols, 
    # contact Yulun Wu at yulunwu8@gmail.com
    
    # Currently SPF is interpolated using the central wavelength. Alternatively,
    # can run 6S and extract the spectually resampled values. 
    
    
    n_angles = 83
    
    ### Columns 
    
    # From 6S code and manual 
    wls = np.array([0.350,0.400,0.412,0.443,0.470,0.488,0.515,0.550,0.590,0.633,0.670,0.694,0.760,0.860,1.240,1.536,1.650,1.950,2.250,3.750]) * 1000
    
    ### Rows, the angles
    
    angles = np.empty(83, dtype=float)
    
    angles[0] = 0.
    angles[1:41] = np.arange(0.75, 40, 1) / 80.5 * 180
    angles[41]= 90.
    angles[42:82] = np.arange(40.25 + 0.5, 80, 1) / 80.5 * 180 # 40.25 is right in the middle
    angles[82]=180.
    
    # reverse the angles 
    angles[::-1].sort()
    
    ###
    

    file_aerosolSPF = 'ancillary/aerosolSPF/' + str(aerosol_type) + '.csv'
    file_aerosolSPF = os.path.join(os.path.dirname(__file__), file_aerosolSPF)
    
    aerosolSPF = np.genfromtxt(file_aerosolSPF, delimiter=',')
    
    # np.shape(aerosolSPF)
    # aerosolSPF[0,:]
    # np.interp(wl, wls, aerosolSPF[0,:])
    
    
    aerosolSPF_wl = np.array([np.interp(wl, wls, aerosolSPF[i,:]) for i in range(np.shape(aerosolSPF)[0])])
    aerosolSPF_wl = aerosolSPF_wl[0:n_angles]
    
    
    df = pd.DataFrame({'Angle':angles, 'Value':aerosolSPF_wl}).sort_values('Angle').reset_index()
    
    # df
    
    # df.plot( 'angle','value', kind='scatter', logy=True)
    
    
    # Already normalized data
    
    
    # df.to_csv('testSPF.csv', index=False)
    return df


if __name__=='__main__':
    aerosol_type = 'Maritime'
    # aerosol_type = 'Continental'
    
    wl = 550
    
    test = find_aerosolSPF(aerosol_type,wl)
    print(test)










# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# intersect


import pandas as pd
import numpy as np
import sys

from .tm_geometry import angle_3d, linear_distance
from copy import copy



# new find_atm two scattering OTs for travelling through multiple layers 
def find_atm2(atm_profile,q1):
    
    z = q1[2]/1000
    
    if z < 0:
        print("ERROR: z < 0")
        return None
    
    # print(atm_profile)
    
    alt_top_minus_z = atm_profile[:,1] - z
    # print('alt_top_minus_z: ' + str(alt_top_minus_z))
    
    
    # index of minimum positive difference 
    idx = np.where(alt_top_minus_z >= 0, alt_top_minus_z, np.inf).argmin()
    # print('Index: ' + str(idx))
    
    
    # print(atm_profile[idx,:])
    
    ot_rayleigh = atm_profile[idx,3]
    ot_mie = atm_profile[idx,4]
    
    return ot_rayleigh, ot_mie
    




# Testing intersecting triangles 
# 2: test boxes first 
# This one is used in the code because it is a lot faster 

def intersect_line_DEMtri2(q0, q1, DEM_tri, print_on = False):
    '''

    Parameters
    ----------
    q0 : list
        starting point of the line.
    q1 : list
        ending point.
    DEM_tri : list
        a list of numpy arrays.

    Returns
    -------
    intersect_tri : pandas dataframe
        collision coordinates, direction normal to the surface
        and linear distance to the starting point.

    '''
    
    # Manual switch 
    print_on = False 
    
    
    ### Identify boxes intersecting the line first     

    tri_x = DEM_tri[0][:,0] 
    tri_y = DEM_tri[0][:,1] 
    tri_z = DEM_tri[0][:,2]
    
    
    tri_x_min = np.min(tri_x,axis=0)
    tri_x_max = np.max(tri_x,axis=0)

    tri_y_min = np.min(tri_y,axis=0)  
    tri_y_max = np.max(tri_y,axis=0)    
    
    
    # Find the box within max z and min z, project it onto XY and select the pixels 
       
    pt_direction_c = [q1[0]-q0[0],q1[1]-q0[1],q1[2]-q0[2]]
    
    z_max = np.max(tri_z) + 0.1 
    z_min = np.min(tri_z) - 0.1
    
    pt_direction_c = q1 - q0
    
    # make box 
    q_z_max = q0 + ( (z_max - q0[2]) / pt_direction_c[2] ) * pt_direction_c
    q_z_min = q0 + ( (z_min - q0[2]) / pt_direction_c[2] ) * pt_direction_c
    
    # First treat all triangles as boxes (max/min XYZ)!!!
    
    box_x_min = min(q_z_max[0],q_z_min[0])
    box_x_max = max(q_z_max[0],q_z_min[0])
    box_y_min = min(q_z_max[1],q_z_min[1])
    box_y_max = max(q_z_max[1],q_z_min[1])
    
    # boxes that cross z
    crossing_z = np.logical_and(np.logical_and(tri_x_min < box_x_max, tri_x_max > box_x_min),
                                np.logical_and(tri_y_min < box_y_max, tri_y_max > box_y_min))
    
    # crossing is for all xyz 
    
    if np.sum(crossing_z) <= 1:
        crossing = crossing_z
        
    else:
    
        if q0[0] == q1[0] and q0[1] == q1[1]:
            if print_on: print('Vertical move')
            crossing = crossing_z
            
        elif q0[0] == q1[0]:
            if print_on: print('Equal x, move along y')
            crossing_y = np.logical_and(tri_x_min < q0[0], tri_x_max > q0[0])
            crossing = np.logical_and(crossing_z, crossing_y)
            
            
        elif q0[1] == q1[1]:
            if print_on: print('Equal y, move along x')    
            crossing_x = np.logical_and(tri_y_min < q0[1], tri_y_max > q0[1])
            crossing = np.logical_and(crossing_z, crossing_x)
            
            
        else:
            if print_on: print('Move along both x and y')
            
            # test if intersects the top or bottom of the square
            
            # equation on the x-y plane
            slope = (q1[1] - q0[1]) / (q1[0] - q0[0])
            y_itcp = q1[1] - q1[0]*slope
            
            
            # crossing the upper boundary (more negative y)
            x_itcp_y_min = (tri_y_min - y_itcp) / slope
            
            # crossing the lower boundary (more positive y)
            x_itcp_y_max = (tri_y_max - y_itcp) / slope
            
            y_itcp_x_min = tri_x_min * slope + y_itcp
            y_itcp_x_max = tri_x_max * slope + y_itcp

            
            crossing_upper = np.logical_and(tri_x_min < x_itcp_y_min, x_itcp_y_min < tri_x_max)
            crossing_lower = np.logical_and(tri_x_min < x_itcp_y_max, x_itcp_y_max < tri_x_max)
            crossing_x = np.logical_or(crossing_upper,crossing_lower) 
            
            
            crossing_left = np.logical_and(tri_y_min < y_itcp_x_min, y_itcp_x_min < tri_y_max)
            crossing_right = np.logical_and(tri_y_min < y_itcp_x_max, y_itcp_x_min < y_itcp_x_max)
            crossing_y = np.logical_or(crossing_left,crossing_right) 
            
            
            crossing_xy = np.logical_or(crossing_x,crossing_y) 
            crossing = np.logical_and(crossing_z, crossing_xy)
            
            
            # np.where(crossing_z)
            # np.where(crossing_xy)
            
            

    # intersecting triangles 
    intersect_tri = pd.DataFrame() 
    
    for tri in DEM_tri: # tri is a set of triangle, ref_tri1[point0-2, xyz:0-2, row, column]
    
        # tri[:,:,crossing]
        for i in range(0,tri[:,:,crossing].shape[2]):

            # print (i)
            
            p0 = tri[:,:,crossing] [0,:,i]
            p1 = tri[:,:,crossing] [1,:,i]
            p2 = tri[:,:,crossing] [2,:,i]
                
            # Extract coordinates of intersection if exists 
            intersect = _intersect_line_triangle(q0,q1,p0,p1,p2)

            
            if (intersect is not None):             
                
                if print_on:
                    print("Line intersecting triangle at: " + str(intersect))
                
                N = np.cross(p1-p0, p2-p0) 
                
                # test normal if in the same direction as the incoming line 
                # if not, reverse normal direction 
                
                if angle_3d(q0,intersect,(intersect + N)) > 90: N = -N
                
                intersect_tri_temp = pd.DataFrame({
                    # coordinates and direction 
                    'X': [intersect[0]],
                    'Y': [intersect[1]],
                    'Z': [intersect[2]],
                    'N_X': [N[0]],
                    'N_Y': [N[1]],
                    'N_Z': [N[2]]
                    })
                
                intersect_tri_temp['linear_distance'] = linear_distance(
                    intersect_tri_temp.iloc[0,0:3].tolist(), 
                    q0)
                
                # intersect_tri = intersect_tri.append(intersect_tri_temp) # append to concat on Feb 4, 2022
                intersect_tri = pd.concat([intersect_tri,intersect_tri_temp])
      
    intersect_tri.reset_index(drop=True, inplace=True)
    return intersect_tri





def _intersect_line_triangle(q1,q2,p1,p2,p3):
    '''

    Parameters
    ----------
    q1 : list
        Coordinates of one end of the line: X, Y, Z.
    q2 : list
        Coordinates of the other end of the line.
    p1 : list
        Coordinates of the first angle of the triangle.
    p2 : list
        Coordinates of the second angle of the triangle.
    p3 : list
        Coordinates of the third angle of the triangle.

    Returns
    -------
    Coordinates of the intersection if they intersect, or 
    None if they don't.


    '''
    # from https://stackoverflow.com/questions/42740765/intersection-between-line-and-triangle-in-3d
    def signed_tetra_volume(a,b,c,d):
        return np.sign(np.dot(np.cross(b-a,c-a),d-a)/6.0)

    s1 = signed_tetra_volume(q1,p1,p2,p3)
    s2 = signed_tetra_volume(q2,p1,p2,p3)

    if s1 != s2:
        s3 = signed_tetra_volume(q1,q2,p1,p2)
        s4 = signed_tetra_volume(q1,q2,p2,p3)
        s5 = signed_tetra_volume(q1,q2,p3,p1)
        if s3 == s4 and s4 == s5:
            n = np.cross(p2-p1,p3-p1)
            t = np.dot(p1-q1,n) / np.dot(q2-q1,n)
            return q1 + t * (q2-q1)
    return None





def reflectance_intersect(q_collision, image_reflectance, cell_size, bg_ref, bg_coords):
    '''
    Find the reflectance of a triangle when collision happens
    June 10, 2021: also if is water 

    Parameters
    ----------
    q_collision : TYPE
        DESCRIPTION.
    image_reflectance : TYPE
        DESCRIPTION.
    cell_size : TYPE
        DESCRIPTION.
    b1_ref : TYPE
        background reflectance 1.
    b2_ref : TYPE
        background reflectance 2.
    b1_a : TYPE
        first point dividing b1_ref and b2_ref.
    b1_b : TYPE
        2nd point.

    Returns
    -------
    triangle_reflectance : TYPE
        DESCRIPTION.

    '''
    
    b1_ref = bg_ref[0]
    b2_ref = bg_ref[1]
    
    image_shape = image_reflectance.shape

    b1_a, b1_b = np.array(bg_coords[0]), np.array(bg_coords[1])
    
    triangle_y = int(q_collision[1]//cell_size)
    triangle_x = int(q_collision[0]//cell_size)
         
    line = b1_a - b1_b
    slope = line[1]/line[0]  # y = slope * x + x_intercept 
    x_intercept = b1_a[1] - slope *b1_a[0]   
    
    # if on the padding triangles 
    
    if (triangle_y < 0 or triangle_x < 0 or 
        triangle_y >= image_shape[0] or triangle_x >= image_shape[1]):
    
        above_line = q_collision[1] >= (q_collision[0] * slope + x_intercept) 
        
        if x_intercept>=0:
            #print('below the line is b1')
            
            if above_line:
                triangle_reflectance = b2_ref
            else:
                triangle_reflectance = b1_ref
                
        else:
            #print('above the line is b1')
            
            if above_line:
                triangle_reflectance = b1_ref
            else:
                triangle_reflectance = b2_ref
        
    else: # if within the original image 
    
        triangle_reflectance =  image_reflectance[triangle_y, triangle_x]
    
    return triangle_reflectance



def reflectance_background(q_collision,bg_ref, bg_coords):
    # Find reflectance when outside the triangles 
    # June 10, 2021: also if is water 
    
    b1_ref = bg_ref[0]
    b2_ref = bg_ref[1]
    
    b1_a, b1_b = np.array(bg_coords[0]), np.array(bg_coords[1])
    line = b1_a - b1_b
    slope = line[1]/line[0]  # y = slope * x + x_intercept 
    x_intercept = b1_a[1] - slope *b1_a[0]  
    
    above_line = q_collision[1] >= (q_collision[0] * slope + x_intercept) 
        
    if x_intercept>=0:
        #print('below the line is b1')
        
        if above_line:
            triangle_reflectance = b2_ref
        else:
            triangle_reflectance = b1_ref
            
    else:
        #print('above the line is b1')
        
        if above_line:
            triangle_reflectance = b1_ref
        else:
            triangle_reflectance = b2_ref
    return triangle_reflectance


def intersect_background(q0,q1,bg_elevation):
    '''
    Find the XYZ of q0-q1 at Z 

    '''

    ratio = (q0[2]-bg_elevation) / (q0[2]-q1[2])
    
    x = q0[0] - ((q0[0]-q1[0])) * ratio
    y = q0[1] - ((q0[1]-q1[1])) * ratio

    return [x,y,bg_elevation]


# test if line intersects atm boundaries 
def intersect_line_boundary(q0,q1,atm_profile,print_on=False):
    '''
    

    return the closest layer 


    '''
    
    '''
    ### testing 
    q0 = [0,0,110_000]
    q1 = [0,0,40_000]
    q0 = [0,0,90_000]
    q1 = [0,0,110_000]    
    ###
    '''
    
    z0, z1 = q0[2]/1000,q1[2]/1000
    
    if z0==z1:
        sys.exit("intersect_line_boundary: z0==z1")
    
    going_up = z1>z0
    # print(going_up)
    
    cross_b = False # if cross a boundary 
    out = False # if out
    
    # output 
    intersect_b = pd.DataFrame() # default output if no crossing boundary 
    
    
    if going_up:
        diff = atm_profile.Alt_bottom - z0
        
        # print('======')
        # print('z0: ' + str(z0))
        # print('atm_profile.Alt_bottom: ' + str(atm_profile.Alt_bottom))
        # print('diff: ' + str(diff))

        # print('diff>0: ' + str(diff>0))
        
        # print('any: ' + str(any(diff>0)))
        
        if any(diff>0): # at least one alt_bottom above z0
            layer_idx = diff[diff>0].idxmin()
            
            # print(layer_idx)
            
            layer_bottom = atm_profile.Alt_bottom[layer_idx]
            # print(layer_bottom)
            
            if (z1 >= layer_bottom):
                # append info 
                cross_b = True 
                altitude = layer_bottom # altitude of crossing boundary 
        else: # all alt_bottom below z0 
            if (z1>=100 and going_up):
                out = True 
                cross_b = True
                altitude = 100
                layer_idx = atm_profile.Alt_bottom.idxmax() # return top layer because OTs are not used 
            
    
    
    else: # going down 
        diff = z0 - atm_profile.Alt_top
        # print('======')
        # print('z0: ' + str(z0))
        # print('atm_profile.Alt_top: ' + str(atm_profile.Alt_top))
        # print('diff: ' + str(diff))

        # print('diff>0: ' + str(diff>0))
        
        # print('any: ' + str(any(diff>0)))
        
        if any(diff>0): # no alt_top above q0 --> no need to edit atm 

            layer_idx = diff[diff>0].idxmin()
            
            # print(layer_idx)
            
            layer_top = atm_profile.Alt_top[layer_idx]
            # print(layer_top)
            
            if(z1 <= layer_top):
                # append info 
                cross_b = True 
                altitude = layer_top # altitude of crossing boundary 
            
    if cross_b: 
        intersect_b = copy(atm_profile.iloc[layer_idx])
        intersect_b['out'] = False
        
        # where crossing boundary  
        xyz = intersect_background(q0,q1,altitude*1000) + [altitude*1000]
        
        intersect_b['X'] = xyz[0]
        intersect_b['Y'] = xyz[1]
        intersect_b['Z'] = xyz[2]
        
        intersect_b['linear_distance'] = linear_distance(q0,xyz)

        
        if out:
            intersect_b['out'] = True
    
    return intersect_b 







'''
# Not used anymore, travel layer by layer 
def find_atm(q0, pt_direction, atm_profile_wl, print_on = False): 
    # Find a pohton in the atmosphere, retrieve OTs 
    
    # Test if the photon alt is equal to any of the upper boundaries 
    equal_top = np.array( q0[2]/1000 == atm_profile_wl.Alt_top )
      
    # At boundary 
    if sum(equal_top): 
    
        atm_index = np.where(equal_top )[0].astype(int)[0]
        
        # Look at moving direction, if moving up, choose the layer above 
        if pt_direction[0]<90:  
            atm_index = atm_index + 1
        elif pt_direction[0]==90: 
            print("\n====== WARNING: photon cannot travel horizontally at a boundary ======")  
            atm_index = None
        
        if print_on:
            print('\nPhoton alt at a boundary, atm_index: ' + str(atm_index))
        
        ot_abs = float(atm_profile_wl.iloc[atm_index].ot_abs)
        ot_rayleigh = float(atm_profile_wl.iloc[atm_index].ot_rayleigh)
        ot_mie = float(atm_profile_wl.iloc[atm_index].ot_mie)        
        
        
    # Not at boundary   
    else:
        
        above_bottom = np.array( q0[2]/1000 > atm_profile_wl.Alt_bottom )
        below_top = np.array( q0[2]/1000 < atm_profile_wl.Alt_top )    
        
        above_or_below = above_bottom & below_top
        
        # if print_on:
        #     print('\n Above_bottom: ')
        #     print(above_bottom)
        #     print('\n Below_top: ')
        #     print(below_top)
        #     print('\n Above_or_below: ')
        #     print(above_or_below)

        # Within a layer 
        if sum(above_or_below): 
            
            atm_index = np.where(above_bottom & below_top)[0].astype(int)[0]
            
            if print_on:
                print('\nPhoton alt within a layer, atm_index: ' + str(atm_index))
            
            ot_abs = float(atm_profile_wl.iloc[atm_index].ot_abs)
            ot_rayleigh = float(atm_profile_wl.iloc[atm_index].ot_rayleigh)
            ot_mie = float(atm_profile_wl.iloc[atm_index].ot_mie)

        # Not at boundary nor within a layer 
        else: 
            # Above all layers 
            if q0[2]/1000 > np.max(atm_profile_wl.Alt_top):
                
                if print_on:
                    print ("\nPhoton alt is above max, setting OTs as 0 ")
                    
                ot_abs, ot_rayleigh, ot_mie = 0, 0, 0
            
            # Unknown 
            else:
                print("====== WARNING: error locating photon in the atm ======")  
                
    if print_on:
        print("ot_abs: {:.2e}, ot_rayleigh: {:.2e}, ot_mie: {:.2e}".format(ot_abs, ot_rayleigh, ot_mie))
    
    return ot_abs, ot_rayleigh, ot_mie
'''





'''
# Testing intersecting triangles, this function is not used 
def intersect_line_DEMtri(q0, q1, DEM_tri, print_on = False): # not used 

    # Parameters
    # ----------
    # q0 : list
    #     starting point of the line.
    # q1 : list
    #     ending point.
    # DEM_tri : list
    #     a list of numpy arrays.

    # Returns
    # -------
    # intersect_tri : pandas dataframe
    #     collision coordinates, direction normal to the surface
    #     and linear distance to the starting point.
    
    # intersecting triangles 
    intersect_tri = pd.DataFrame() 
    
    for tri in DEM_tri: 
    
        for row in range (0, tri.shape[2]):
            for col in range (0, tri.shape[3]):
                #print (row, col)
        
                p0 = tri[0,:,row,col] 
                p1 = tri[1,:,row,col]
                p2 = tri[2,:,row,col]
                
                # Extract coordinates of intersection if exists 
                intersect = _intersect_line_triangle(q0,q1,p0,p1,p2)
                if (intersect is not None): 
                    
                    
                    if print_on:
                        print("Line intersecting triangle at: " + str(intersect))
                    
                    N = np.cross(p1-p0, p2-p0) #maybe faster than the line above 
                    

                    if angle_3d(q0,intersect,N) > 90: N = -N
                    
                    intersect_tri_temp = pd.DataFrame({

                        'X': [intersect[0]],
                        'Y': [intersect[1]],
                        'Z': [intersect[2]],
                        'N_X': [N[0]],
                        'N_Y': [N[1]],
                        'N_Z': [N[2]]
                        })
                    
                    intersect_tri_temp['linear_distance'] = linear_distance(
                        intersect_tri_temp.iloc[0,0:3].tolist(), 
                        q0)
                    
                    # intersect_tri = intersect_tri.append(intersect_tri_temp) # append to concat on Feb 4, 2022
                    intersect_tri = pd.concat([intersect_tri,intersect_tri_temp])
      
    intersect_tri.reset_index(drop=True, inplace=True)
    return intersect_tri
'''












# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.




# Atmosphere object  

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

import Py6S
from Py6S.Params.atmosprofile import AtmosProfile
from Py6S.Params.aeroprofile import AeroProfile

from .Aerosol import find_aerosolSPF

import os.path


class Atmosphere(): 
    '''Create an Atmosphere object that is wavelength independent. I.e., a wavelength-dependent atmosphere is generated every time TMart is run.
    
    Arguments:

    * ``atm_profile`` -- AtmosProfile object from Py6S. 
    * ``aot550`` -- AOT at 550nm.
    * ``aerosol_type`` -- 'BiomassBurning', 'Continental', 'Desert', 'Maritime', 'Stratospheric' or 'Urban', as provided by 6S. 
    * ``wl`` -- central wavelength in nm.
    * ``n_layers`` -- Number of atmosphere layers to use. Default 20. 
    * ``AEROSOL_SCALE_HEIGHT`` -- Aerosol scale height in km. Default 2km. 
    * ``no_absorption`` -- Boolean, if yes -> remove all absorption. 
    * ``specify_ot_rayleigh`` -- Specify rayleigh optical thickness, only for testing.
    * ``specify_abs`` -- Specifiy absorption optical thickness, only for testing. 
    

    Example usage:: 
    
      from Py6S.Params.atmosprofile import AtmosProfile

      # Atmophere profile comes from 6S
      atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
      aerosol_type = 'Maritime' 
      aot550 = 0.1
      n_layers = 20
      aerosol_scale_height = 2 # Unless you have a reason, don't change this
        
      # Synthesize an atmosphere object    
      my_atm = tmart.Atmosphere(atm_profile, aot550, aerosol_type, n_layers, aerosol_scale_height)
    '''

    def __init__(self,atm_profile, aot550 = 0, 
                 aerosol_type='Maritime' , 
                 n_layers=20, AEROSOL_SCALE_HEIGHT=2, no_absorption = False, specify_ot_rayleigh = -1, specify_abs = -1):
        
        self.atm_profile = atm_profile
        self.aot550 = aot550
        
        
        self.aerosol_type = aerosol_type
        
        # Default 20 layers 
        self.n_layers = n_layers
        
        # Default aerosol scale height is 2km
        self.aerosol_scale_height = AEROSOL_SCALE_HEIGHT
     
        # atm is always 100km thick in total, following 6S 
        self.atm_height = 100
        self.layer_height = self.atm_height/self.n_layers

        # special scenarios 
        self.no_absorption = no_absorption
        self.specify_ot_rayleigh = specify_ot_rayleigh
        self.specify_abs = specify_abs
        
    
    ### The main one dealing with wavelength and band 
    
    # return everything at a single wavelength 
    # supports Aerosol LUT AND 6S wavelengths 
    # Return a table of OTs at different heights  +  aerosol_SPF
    def _wavelength(self, wl, band=None): 
        
        self.wl = wl
        
        # Find bottom, mean and top altitudes of layers 
        self.layers_alts_bottom = np.linspace(0, self.atm_height - self.layer_height, self.n_layers)  
        self.layers_alts_mean = np.linspace(self.layer_height/2, self.atm_height - self.layer_height/2, self.n_layers)  
        self.layers_alts_top = np.linspace(self.layer_height, self.atm_height , self.n_layers)  
        
        # Calculate tau in each layer and extract the molecular profile at one wavelength 
        layers_ot_molecule, layers_ot_rayleigh = self._atm_profile_wl(band)
        
        # ot_mie is aerosol scattering, ot_aerosol is aerosol absorption 
        layers_ot_mie, layers_ot_aerosol = self._aerosol_wl(band)

        # Retrieve aerosol SPF
        aerosol_SPF = find_aerosolSPF(self.aerosol_type,wl)
    
        layers_ot_total_abs = layers_ot_molecule + layers_ot_aerosol
        
        # A data frame of optical thickness 
        atm_OT = pd.DataFrame({'Alt_bottom': self.layers_alts_bottom,
                            'Alt_top': self.layers_alts_top,  
                            'ot_abs': layers_ot_total_abs, 
                            'ot_rayleigh': layers_ot_rayleigh,
                            'ot_mie': layers_ot_mie
                            })
        
        if self.specify_ot_rayleigh == -1:
            pass
        else:
            atm_OT['ot_rayleigh'] = self.specify_ot_rayleigh       
        
        
        atm_OT['ot_scatt'] = atm_OT.ot_rayleigh + atm_OT.ot_mie
        atm_OT['l_height'] = atm_OT.Alt_top - atm_OT.Alt_bottom
        atm_OT['percentage'] = 0 # used to calculate travelling distance 
        
        
        if self.no_absorption:
            atm_OT['ot_abs'] = 0
            
        if self.specify_abs == -1:
            pass
        else:
            atm_OT['ot_abs'] = self.specify_abs               
        
        return atm_OT, aerosol_SPF
    
    
    
    # Extract the molecular profile at one wavelength 
    # Return two lists: ot_molecule and ot_rayleigh 
    def _atm_profile_wl(self,band): 

        layers_ot_molecule = []
        layers_ot_rayleigh = []
        
        layers_alts_bottom = self.layers_alts_bottom
        layers_alts_top = self.layers_alts_top
           
        
        for i_layer in range(len(layers_alts_bottom)):
        
            alt_bottom = 0
            alt_top = layers_alts_top[i_layer]
            
            # Initiate an SixS object 
            s = Py6S.SixS()
            
            # Wavelength and band 
            s.wavelength = Py6S.Wavelength(self.wl/1000)   
            if band is not None:
                s.wavelength = band
            
            # Other parameters 
            s.atmos_profile = self.atm_profile        
            s.geometry = Py6S.Geometry.User()
            s.geometry.solar_z, s.geometry.solar_a = 0, 0
            s.geometry.view_z, s.geometry.view_a = 0, 0
            s.aero_profile = AeroProfile.PredefinedType(AeroProfile.NoAerosols)
            
            # Altitudes 
            s.altitudes.set_target_custom_altitude(alt_bottom)
            
            if alt_top < 100:
                s.altitudes.set_sensor_custom_altitude(alt_top)
            else:
                s.altitudes.set_sensor_satellite_level()    
                s.altitudes.set_sensor_custom_altitude(99.99999)
        
            s.run()
            
            # print(s.outputs.fulltext)
            # print(s.outputs.optical_depth_total)
            # print(s.outputs.optical_depth_total.rayleigh)
    
            tao_rayleigh = s.outputs.optical_depth_plane.rayleigh
    
            # Molecular absorption 
            T_molecule = s.outputs.transmittance_global_gas.upward 
            tao_molecule = -np.log(T_molecule)
            
            # we keep last tao_molecule and tao_rayleigh for full atmosphere 
            layers_ot_molecule.append(tao_molecule)
            layers_ot_rayleigh.append(tao_rayleigh)
                
        
        # Convert 0-10, 0-20... to 0-10, 10-20...
        layers_ot_molecule_new = [layers_ot_molecule[0]]
        layers_ot_rayleigh_new = [layers_ot_rayleigh[0]]    
        
        for i_layer in range(1, len(layers_ot_molecule)):
            
            i_molecule = layers_ot_molecule[i_layer] - layers_ot_molecule[i_layer-1] 

            if i_molecule < 0:
                i_molecule = 0
            
            layers_ot_molecule_new.append(i_molecule)
            
            i_rayleigh = layers_ot_rayleigh[i_layer] - layers_ot_rayleigh[i_layer-1] 

            if i_rayleigh < 0:
                i_rayleigh = 0
            
            layers_ot_rayleigh_new.append(i_rayleigh)            
            
        # percentage in each layer, redistribute total  
        if sum(layers_ot_molecule_new)> 0: # no action in case of full transparency 
            relative_mol = layers_ot_molecule_new/np.sum(layers_ot_molecule_new)
            layers_ot_molecule_new = tao_molecule * relative_mol  
        if sum(layers_ot_rayleigh_new)> 0: 
            relative_ray = layers_ot_rayleigh_new/np.sum(layers_ot_rayleigh_new)    
            layers_ot_rayleigh_new = tao_rayleigh * relative_ray


        return layers_ot_molecule_new, layers_ot_rayleigh_new   
    
    


    # Extract aerosol profile at one wavelength 
    # Find the spectral dependence of AOT and then apply it to AOT550
    def _aerosol_wl(self, band):
        
        # Initiate an SixS object
        s = Py6S.SixS()
        
        # Band and wavelength 
        s.wavelength = Py6S.Wavelength(self.wl/1000)   
        if band is not None:
            s.wavelength = band
        
        # Other parameters 
        s.atmos_profile = self.atm_profile        
        s.geometry = Py6S.Geometry.User()
        s.geometry.solar_z, s.geometry.solar_a = 0, 0
        s.geometry.view_z, s.geometry.view_a = 0, 0
        s.altitudes.set_target_custom_altitude(0)
        s.altitudes.set_sensor_satellite_level()
        
        aerosol_dict = {    "NoAerosols": 0,
                        "Continental" : 1,
                        "Maritime" : 2,
                        "Urban" : 3,
                        "Desert" : 5,
                        "BiomassBurning" : 6,
                        "Stratospheric" : 7}
        
        
        s.aero_profile = AeroProfile.PredefinedType(aerosol_dict[self.aerosol_type])
        s.aot550 = 1
        s.run()
        
    
        aerosol_EXT_r550 = s.outputs.optical_depth_total.aerosol # ratio, relative to 550
        
        aerosol_EXT = self.aot550 * aerosol_EXT_r550 # total extinction 
        aerosol_SSA = s.outputs.single_scattering_albedo.aerosol # single scattering albedo 
        
        ot_mie = aerosol_EXT * aerosol_SSA # scattering by aerosols 
        ot_aerosol = aerosol_EXT - ot_mie  # absorption by aerosols 
        
        # Concentration at mean altitudes, make integrated area???
        conc_relative = np.exp(-self.layers_alts_mean/self.aerosol_scale_height)
        conc_normalized = conc_relative / np.sum(conc_relative)        

        # Divide optical thickness into layers, weighted by concentration at mean heights 
        layers_ot_mie = ot_mie * conc_normalized
        layers_ot_aerosol = ot_aerosol * conc_normalized # aerosol absorption
  
        return layers_ot_mie, layers_ot_aerosol # scattering and absorption by aerosol, respectively 














# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.





from .Surface import Surface, SpectralSurface
from .Atmosphere import Atmosphere

# from .tm import Tmart

from .Tmart import Tmart


from .tm_calcref import calc_ref 

from .tm_geometry import dirP_to_coord















# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Surface object

import numpy as np
import pandas as pd

from scipy.interpolate import interp1d

import os.path


class SpectralSurface():
    '''Create an object to capture the spectral reflectance of surfaces when looping wavelengths. This can be used as input to reflectance in the Surface object.
    
    Arguments:

    * ``land_cover`` -- Text, currently support 'soil', 'vegetation', 'water' and 'water_chl1' ([chla]=1).
    

    Example usage::

      # Create object
      water = tmart.SpectralSurface('water_chl1')
      
      # Find spectral reflectance at 400nm
      water.wl(400)
      
      # Create a spectral surface in a numpy array 
      wl = 400 # your variable in a loop
      np.full((2, 2), water.wl(wl))
      

    '''
    def __init__(self,land_cover):
        
        file_name = os.path.join(os.path.dirname(__file__), 'ancillary', str(land_cover) + '.csv')
        
        try:
        
        
            df = pd.read_csv(file_name)
            
            wavelength = df.wl.to_numpy() * 1000
            value = df.value.to_numpy() /100
            
            self.f = interp1d(wavelength, value) # , kind='cubic')
        
        except: 
            print('Warning: failed to open ' + str(file_name))

        
    def wl(self,wavelength):
        '''
        Specify the wavelength of interest in nm and returns the reflectance of the surface. 
        '''
        
        reflectance = self.f(wavelength).item()
        if reflectance < 0: reflectance = 0
        return reflectance 





class Surface(): 
    '''Create an Surface object. 
    
    Arguments:

    * ``DEM`` -- Numpy array, Digital Elevation Model, the elevation of pixels.
    * ``reflectance`` -- Numpy array, reflectance of land or water-leaving reflectance of water, Lambertian.
    * ``isWater`` -- Numpy array, which pixels are water pixels. 1 is water, 0 is land.
    * ``cell_size`` -- The width and length of each pixel, in meters.
    * ``alignPixels`` -- Boolean. If true: the southeast corner of the pixel will take the elevation in DEM. If false: the centre of the pixel will take the elevation in DEM.

    

    Example usage::

      image_DEM = np.array([[0,0],[0,0]]) # in meters
      image_reflectance = np.array([[0.1,0.1],[0.1,0.1]]) # unitless     
      image_isWater = np.array([[1,1],[1,1]]) # 1 is water, 0 is land
      cell_size = 20_000 
      my_surface = tmart.Surface(image_DEM,image_reflectance,image_isWater,cell_size)  

    '''
    
    def __init__(self,DEM,reflectance,isWater,cell_size,alignPixels=True):
        self.DEM = DEM
        self.reflectance = reflectance
        self.isWater = isWater
        self.cell_size = cell_size
        self.alignPixels = alignPixels
        
        
        if self.DEM.shape != self.reflectance.shape:
            print('WARNING: DEM and reflectance images do not have the same shape')

        # Background parameters that can be modified using object.set_background function
        self.bg_ref = None
        self.bg_isWater = None
        self.bg_elevation = None         
        self.bg_coords = None

        # Triangulated DEM, two sets of three dimensional triangles 
        self.DEM_triangulated = None
        
        self.set_background() # test March 12, 2022
        self._triangulate_DEM()
        
        self.x_min = np.min(self.DEM_triangulated[0][:,0,:,:])
        self.x_max = np.max(self.DEM_triangulated[0][:,0,:,:])
        self.y_min = np.min(self.DEM_triangulated[0][:,1,:,:])
        self.y_max = np.max(self.DEM_triangulated[0][:,1,:,:])
        
        
        
        
    # Two Reflectance Surfaces and if water 
    def set_background(self,bg_ref=None, bg_isWater=None, bg_elevation=None, bg_coords=None):
        '''Set background information, 1 or 2 background surfaces can be set;
        If 2 surfaces: the first background is the one closer to [0,0]
        
        Arguments:

        * ``bg_ref`` -- A number or a list of two numbers. Reflectance of each background surface. Default average reflectance of the pixels.
        * ``bg_isWater`` -- A number or a list of two numbers. If each of the background surface is water. Default land.
        * ``bg_elevation`` -- A number. The elevation of the background surfaces. Default 0.
        * ``bg_coords`` -- A list of two lists of two numbers. Two XY coordinates divide the two background surfaces. Default [[0,0],[1,1]].


        Example usage::

          my_surface.set_background(bg_ref        = [0.02,0.02], # background reflectance
                                    bg_isWater    = [0,0], # if is water
                                    bg_elevation  = 0, # elevation of both background
                                    bg_coords     = [[0,0],[10,10]]) # a line dividing two background 

        '''

        # default: average reflectance, elevation 0, no water 
  
        
        # Reflectances of 2 background surfaces, first surface close to [0,0]
        if bg_ref==None: # default, average reflectance 
            self.bg_ref = [np.average(self.reflectance),np.average(self.reflectance)]  
        elif isinstance(bg_ref,list): # if list, take it 
            self.bg_ref = bg_ref
        else: # integer or float 
            self.bg_ref = [bg_ref,bg_ref]
        
        # isWater list 
        if bg_isWater==None: # default, not water 
            self.bg_isWater = [0,0]
        elif isinstance(bg_isWater,list): # if list, take it 
            self.bg_isWater = bg_isWater
        else: # integer or float 
            self.bg_isWater = [bg_isWater,bg_isWater]
        
        
        # Elevation of the background surfaces, has to be the same 
        if bg_elevation==None:
            self.bg_elevation = 0
        else:
            self.bg_elevation = bg_elevation
        
        # Two points dividing 2 background reflecting surfaces 
        if bg_coords==None: # Any value will do 
            self.bg_coords = np.array([[0,0],[1,1]])      
        else:
            self.bg_coords = np.array(bg_coords).astype(np.float32)
            # two X's can't be the same
            if self.bg_coords[0,0] == self.bg_coords[1,0]:
                self.bg_coords[1,0] = self.bg_coords[1,0] + 0.01            

        self._triangulate_DEM()


    def _triangulate_DEM(self):
        
        '''
        Make two sets of three dimensional triangles to model the surface. 
        
        Parameters
        ----------
        DEM : np array
            elevation of cells.
        cell_size : number
            width of cells.
        alignPixels: boolean
            if align with pixels (elevation will be xy positive bottom right corner of the pixel)
    
        Returns
        -------
        list
            two np arrays.
            to index the triangles: ref_tri1[point0-2, xyz:0-2, row, column]
    
        '''
        
        alignPixels=self.alignPixels
        
        if self.bg_elevation is None:
            print("====== WARNING: DEM not triangulated because of missing background elevation ======")
        else:
            
            if alignPixels: # pad two layers on left and top 
                DEM = np.pad(self.DEM,((2,1),(2,1)),'constant',constant_values = (self.bg_elevation,self.bg_elevation))
                # one elevation only  
                
            else: # pad one layer on each side 
                DEM = np.pad(self.DEM,((1,1),(1,1)),'constant',constant_values = (self.bg_elevation,self.bg_elevation))
            
            # cell_size 
            CZ = self.cell_size 
            
            ref_tri_p1 = np.array([
                # x
                np.tile(np.array(range(0,DEM.shape[1]-1)) * CZ - CZ/2,(DEM.shape[0]-1,1)),
                # y
                np.transpose([np.array(range(0,DEM.shape[0]-1)) * CZ - CZ/2]*(DEM.shape[1]-1)),
                # z
                DEM[0:DEM.shape[0]-1,0:DEM.shape[1]-1]
                ])
            
            ref_tri_p2 = np.array([
                # x
                np.tile(np.array(range(0,DEM.shape[1]-1)) * CZ - CZ/2,(DEM.shape[0]-1,1)),
                # y
                np.transpose([np.array(range(1,DEM.shape[0])) * CZ - CZ/2]*(DEM.shape[1]-1)),
                # z
                DEM[1:DEM.shape[0],0:DEM.shape[1]-1]
                ])
            
            ref_tri_p3 = np.array([
                # x
                np.tile(np.array(range(1,DEM.shape[1])) * CZ - CZ/2,(DEM.shape[0]-1,1)),
                # y
                np.transpose([np.array(range(1,DEM.shape[0])) * CZ - CZ/2]*(DEM.shape[1]-1)),
                # z
                DEM[1:DEM.shape[0],1:DEM.shape[1]]
                ])
            
            ref_tri_p4 = np.array([
                # x
                np.tile(np.array(range(1,DEM.shape[1])) * CZ - CZ/2,(DEM.shape[0]-1,1)),
                # y
                np.transpose([np.array(range(0,DEM.shape[0]-1)) * CZ - CZ/2]*(DEM.shape[1]-1)),
                # z
                DEM[0:DEM.shape[0]-1,1:DEM.shape[1]]
                ])    
            
            if alignPixels:
                ref_tri_p1[0] = ref_tri_p1[0] - CZ/2
                ref_tri_p1[1] = ref_tri_p1[1] - CZ/2
                
                ref_tri_p2[0] = ref_tri_p2[0] - CZ/2
                ref_tri_p2[1] = ref_tri_p2[1] - CZ/2
                
                ref_tri_p3[0] = ref_tri_p3[0] - CZ/2
                ref_tri_p3[1] = ref_tri_p3[1] - CZ/2
                
                ref_tri_p4[0] = ref_tri_p4[0] - CZ/2
                ref_tri_p4[1] = ref_tri_p4[1] - CZ/2
                
            ref_tri1 = np.array([ref_tri_p1, ref_tri_p2, ref_tri_p3])
            ref_tri2 = np.array([ref_tri_p1, ref_tri_p4, ref_tri_p3])
            
            self.DEM_triangulated = [ref_tri1,ref_tri2]

        















# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Geometry 


import math
import numpy as np
import sys


def rad(degree): 
    return degree/180*math.pi



def dirP_to_coord(distance, direction):
    '''
    convert polar coordinates and distance to coordinates 
    
    
    Parameters
    ----------
    distance : integer or float 
        Vector: the linear distance or speed.
    direction : list of two integer or float 
        First zenith: up positive Z 0, positive Y 90. 0-180
        Second azimuthal: positive X 0, positive Y 90. 0-360

    Returns
    -------
    coords : list
        Vector in 3 dimensions: X, Y, and Z.

    '''
    # Two inputs: vector

    coords = np.array([0,0,0],float)  
    #math.sqrt(coords[0]**2 + coords[1]**2 +coords[2]**2) 
    
    coords[0] = distance * math.sin(rad(direction[0])) * math.cos(rad(direction[1])) # x
    coords[1] = distance * math.sin(rad(direction[0])) * math.sin(rad(direction[1])) # y
    coords[2] = distance * math.cos(rad(direction[0])) #z
    return coords



def angle_3d(a,b,c):
    '''
    
    calculate the angle between 3 points in 3d 
    source: https://stackoverflow.com/questions/35176451/python-code-to-calculate-angle-between-three-point-using-their-3d-coordinates
    
    '''
    
    a,b,c = np.array(a),np.array(b),np.array(c)

    ba = a - b
    bc = c - b
    
    
    if np.linalg.norm(ba)==0:
        print(a)
        print(b)
        print(c)
        sys.exit("error1 angle_3d")
        
    if np.linalg.norm(bc)==0:
        print(a)
        print(b)
        print(c)
        sys.exit("error2 angle_3d")        
    
    
    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
    angle = np.arccos(cosine_angle)
    
    # np.rad2deg(2.1034) # useful numpy operations 
    # np.deg2rad(120.5159)
    
    return np.degrees(angle)



def linear_distance(pt1, pt2):
    '''
    each point should be a list of 3 coordinates: X, Y, Z
 
    '''
    distance = ((pt2[0]-pt1[0])**2 + (pt2[1]-pt1[1])**2 + (pt2[2]-pt1[2])**2)**(1/2)
    return distance


def dirC_to_dirP(xyz): 
    #takes list xyz (single coord), return theta, phi, r 
    x       = xyz[0]
    y       = xyz[1]
    z       = xyz[2]
    r       =  (x*x + y*y + z*z)**0.5
    theta   =  math.acos(z/r)*180/ math.pi #to degrees
    phi     =  math.atan2(y,x)*180/ math.pi
    
    if phi <0:
        phi = phi +360
    
    return [theta,phi,r]



def rotation_matrix(axis, theta): 
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    source: https://stackoverflow.com/questions/6802577/rotation-of-3d-vector
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])




def dirC_to_coord (direction,q0,linear_length): 
    '''
    scale a cartesian direction to coordinates with a starting point and a length 

    '''
    
    linear_distance = ((direction[0]-q0[0])**2 + 
                       (direction[1]-q0[1])**2 + 
                       (direction[2]-q0[2])**2)**(1/2)
    
    
    linear_distance = ((direction[0])**2 + 
                       (direction[1])**2 + 
                       (direction[2])**2)**(1/2)
    
    scale = linear_length / linear_distance
    
    scaled = [q0[0]+(direction[0]) * scale,
              q0[1]+(direction[1]) * scale,
              q0[2]+(direction[2]) * scale]
    
    return scaled






